// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    config::CpuConfig,
    instruction::{MOp, MOpDestReg, MOpRegNum, MOpTrait, PRegNum, RenamedMOp},
    reg_alloc::RetireQueueIndex,
    rename_table::{rename_table, RenameTablePortConfig},
    unit::UnitMOp,
    util::array_vec::{ArrayVec, Length, ReadyValidArray},
};
use fayalite::{
    prelude::*,
    util::{prefix_sum::PrefixSumAlgorithm, ready_valid::ReadyValid},
};

#[hdl]
pub struct RenamedInsnDest<UnitNumWidth: Size, OutRegNumWidth: Size, RetireQueueIndexWidth: Size> {
    pub retire_queue_index: RetireQueueIndex<RetireQueueIndexWidth>,
    pub p_reg_num: PRegNum<UnitNumWidth, OutRegNumWidth>,
}

impl CpuConfig {
    pub fn renamed_insn_dest(&self) -> RenamedInsnDest<DynSize, DynSize, DynSize> {
        RenamedInsnDest[self.unit_num_width()][self.out_reg_num_width]
            [self.retire_queue_index_width()]
    }
    pub fn renamed_mop_in_unit(
        &self,
    ) -> RenamedMOp<RenamedInsnDest<DynSize, DynSize, DynSize>, DynSize> {
        RenamedMOp[self.renamed_insn_dest()][self.p_reg_num_width()]
    }
}

#[hdl]
pub struct InstructionRenameInputInsn<UnitNumWidth: Size, OutRegNumWidth: Size> {
    pub mop: MOp,
    pub renamed_dest: PRegNum<UnitNumWidth, OutRegNumWidth>,
}

impl CpuConfig {
    pub fn instruction_rename_input_insn(&self) -> InstructionRenameInputInsn<DynSize, DynSize> {
        InstructionRenameInputInsn[self.unit_num_width()][self.out_reg_num_width]
    }
}

#[hdl]
struct InsnsInPrefixSummary<FetchWidth: Size> {
    all_ready: Bool,
    ready_count: Length<FetchWidth>,
    retire_queue_used: Length<FetchWidth>,
}

#[hdl_module]
pub fn instruction_rename(config: &CpuConfig) {
    #[hdl]
    let cd: ClockDomain = m.input();
    #[hdl]
    let insns_in: ReadyValidArray<InstructionRenameInputInsn<DynSize, DynSize>, DynSize> =
        m.input(ReadyValidArray[config.instruction_rename_input_insn()][config.fetch_width.get()]);
    #[hdl]
    let start_retire_queue_index: RetireQueueIndex<DynSize> = m.input(config.retire_queue_index());
    #[hdl]
    let end_retire_queue_index: RetireQueueIndex<DynSize> = m.output(config.retire_queue_index());
    #[hdl]
    let insns_out: Array<
        ReadyValid<RenamedMOp<RenamedInsnDest<DynSize, DynSize, DynSize>, DynSize>>,
    > = m.output(
        Array[ReadyValid[RenamedMOp[config.renamed_insn_dest()][config.p_reg_num_width()]]]
            [config.fetch_width.get()],
    );

    // TODO: handle resetting table after cancelling instructions

    #[hdl]
    let insns_ready_or_move = wire(Array[Bool][config.fetch_width.get()]);

    for (insn_ready_or_move, insn_out) in insns_ready_or_move.into_iter().zip(insns_out) {
        connect(insn_ready_or_move, insn_out.ready);
    }

    ArrayVec::for_each(insns_in.data, |fetch_index, input_insn| {
        #[hdl]
        match input_insn.mop {
            UnitMOp::<_, _, _>::TransformedMove(_) => {
                connect(insns_ready_or_move[fetch_index], true);
            }
            UnitMOp::<_, _, _>::AluBranch(_) | UnitMOp::<_, _, _>::LoadStore(_) => {}
        }
    });

    let insns_in_prefix_summary_ty = InsnsInPrefixSummary[config.fetch_width.get()];
    #[hdl]
    let insns_in_prefix_summaries =
        wire(Array[insns_in_prefix_summary_ty][config.fetch_width.get()]);
    let insns_in_prefix_summaries_vec = PrefixSumAlgorithm::WorkEfficient.run(
        (0..config.fetch_width.get()).map(|fetch_index| {
            #[hdl]
            let insns_in_prefix_summary_in = wire(insns_in_prefix_summary_ty);
            #[hdl]
            let InsnsInPrefixSummary::<_> {
                all_ready,
                ready_count,
                retire_queue_used,
            } = insns_in_prefix_summary_in;
            connect(all_ready, insns_out[fetch_index].ready);
            connect(
                ready_count,
                Expr::ty(ready_count).cast_from_uint_unchecked(all_ready.cast_to(UInt[1])),
            );
            connect(retire_queue_used, Expr::ty(retire_queue_used).zero());
            #[hdl]
            if let HdlSome(input_insn) = ArrayVec::get(insns_in.data, fetch_index) {
                connect(retire_queue_used, ready_count);
                #[hdl]
                match input_insn.mop {
                    UnitMOp::<_, _, _>::TransformedMove(_) => {
                        connect(all_ready, true);
                        connect(retire_queue_used, Expr::ty(retire_queue_used).zero());
                    }
                    UnitMOp::<_, _, _>::AluBranch(_) | UnitMOp::<_, _, _>::LoadStore(_) => {}
                }
            }
            insns_in_prefix_summary_in
        }),
        |l, r| {
            #[hdl]
            let insns_in_prefix_summary_merge = wire(insns_in_prefix_summary_ty);
            #[hdl]
            let InsnsInPrefixSummary::<_> {
                all_ready,
                ready_count,
                retire_queue_used,
            } = insns_in_prefix_summary_merge;
            connect(all_ready, l.all_ready & r.all_ready);
            #[hdl]
            if l.all_ready {
                connect(
                    ready_count,
                    Expr::ty(ready_count).cast_from_uint_unchecked(
                        Length::as_uint(l.ready_count) + Length::as_uint(r.ready_count),
                    ),
                );
                connect(
                    retire_queue_used,
                    Expr::ty(retire_queue_used).cast_from_uint_unchecked(
                        Length::as_uint(l.retire_queue_used) + Length::as_uint(r.retire_queue_used),
                    ),
                );
            } else {
                connect(ready_count, l.ready_count);
                connect(retire_queue_used, l.retire_queue_used);
            }
            insns_in_prefix_summary_merge
        },
    );
    for (l, r) in insns_in_prefix_summaries
        .into_iter()
        .zip(insns_in_prefix_summaries_vec)
    {
        connect(l, r);
    }
    connect(
        insns_in.ready,
        insns_in_prefix_summaries[config.fetch_width.get() - 1].ready_count,
    );

    #[hdl]
    let retire_queue_indexes =
        wire(Array[config.retire_queue_index()][config.fetch_width.get() + 1]);
    connect(retire_queue_indexes[0], start_retire_queue_index);
    connect(
        end_retire_queue_index,
        retire_queue_indexes[config.fetch_width.get()],
    );
    for (retire_queue_index, insns_in_prefix_summary) in retire_queue_indexes
        .into_iter()
        .skip(1)
        .zip(insns_in_prefix_summaries)
    {
        connect_any(
            retire_queue_index.index,
            start_retire_queue_index.index
                + Length::as_uint(insns_in_prefix_summary.retire_queue_used),
        );
    }

    let mut port_configs = Vec::new();
    let mut src_reg_count = 0;
    MOpTrait::for_each_src_reg(MOp.uninit(), &mut |_, src_index| {
        src_reg_count = src_reg_count.max(src_index + 1);
    });
    for _ in 0..config.fetch_width.get() {
        for _ in 0..src_reg_count {
            port_configs.push(RenameTablePortConfig::Read {
                addr_range: MOpRegNum::NON_CONST_REG_NUMS,
            });
        }
        for dest_reg_kind in MOpDestReg::REG_KINDS {
            port_configs.push(RenameTablePortConfig::Write {
                addr_range: dest_reg_kind.reg_num_range(),
            });
        }
    }

    #[hdl]
    let rename_table = instance(rename_table(config, &port_configs));

    connect(rename_table.cd, cd);

    for read_port in rename_table.read_ports {
        connect_any(read_port.addr, 0_hdl_u0);
    }
    for write_port in rename_table.write_ports {
        connect_any(write_port.addr, 0_hdl_u0);
        connect_any(write_port.data, config.p_reg_num().const_zero());
    }

    let mut read_port_index = 0;
    let mut write_port_index = 0;
    ArrayVec::for_each(
        ReadyValidArray::firing_data(insns_in),
        |fetch_index, input_insn| {
            #[hdl]
            let InstructionRenameInputInsn::<_, _> { mop, renamed_dest } = input_insn;
            let new_dest = #[hdl]
            RenamedInsnDest::<_, _, _> {
                retire_queue_index: retire_queue_indexes[fetch_index],
                p_reg_num: renamed_dest,
            };
            let insn_out = MOpTrait::map_regs(
                mop,
                new_dest,
                config.p_reg_num_width(),
                &mut |src_reg, src_index| {
                    connect(
                        rename_table.read_ports[read_port_index + src_index].addr,
                        src_reg.cast_bits_to(MOpRegNum),
                    );
                    rename_table.read_ports[read_port_index + src_index]
                        .data
                        .cast_to_bits()
                },
            );
            let insn_out = UnitMOp::try_with_transformed_move_op(
                insn_out,
                config.renamed_mop_in_unit().TransformedMove,
                |v: Expr<HdlOption<_>>, _| connect(v, Expr::ty(v).HdlNone()),
            );
            connect(insns_out[fetch_index].data, insn_out);
            read_port_index += src_reg_count;
            for dest_reg in MOpDestReg::regs(MOpTrait::dest_reg(mop)) {
                connect(rename_table.write_ports[write_port_index].addr, dest_reg);
                connect(
                    rename_table.write_ports[write_port_index].data,
                    renamed_dest,
                );
                write_port_index += 1;
            }
        },
    );
}
